@@ -14,6 +14,7 @@
 #include <memory.h>
 #endif /* HAVE_MEMORY_H */
 #include <sys/types.h>
+#include <sys/stat.h>
 #ifndef HAVE_NET_SOCKET_H
 #include <sys/socket.h>
 #else
@@ -36,9 +37,22 @@
 #else
 #include <varargs.h>
 #endif
+#if TIME_WITH_SYS_TIME
+# include <sys/time.h>
+# include <time.h>
+#else
+# if HAVE_SYS_TIME_H
+#  include <sys/time.h>
+# else
+#  include <time.h>
+# endif
+#endif
+
 #include "socket.h"
 #include "fetchmail.h"
+#include "getaddrinfo.h"
 #include "i18n.h"
+#include "sdump.h"
 
 /* Defines to allow BeOS and Cygwin to play nice... */
 #ifdef __BEOS__
@@ -61,26 +75,15 @@ static ssize_t cygwin_read(int sock, void *buf, size_t count);
 
 /* We need to define h_errno only if it is not already */
 #ifndef h_errno
-
-#ifdef HAVE_RES_SEARCH
-/* some versions of FreeBSD should declare this but don't */
+# if !HAVE_DECL_H_ERRNO
 extern int h_errno;
-#else
-/* pretend we have h_errno to avoid some #ifdef's later */
-static int h_errno;
-#endif
-
+# endif
 #endif /* ndef h_errno */
 
-extern int mailserver_socket_temp;	/* Socket to close if connect timeout */
-
-#if NET_SECURITY
-#include <net/security.h>
-#endif /* NET_SECURITY */
-
 #ifdef HAVE_SOCKETPAIR
-char *const *parse_plugin(const char *plugin, const char *host, const char *service)
-{	const char **argvec;
+static char *const *parse_plugin(const char *plugin, const char *host, const char *service)
+{
+	char **argvec;
 	const char *c, *p;
 	char *cp, *plugin_copy;
 	unsigned int plugin_copy_len;
@@ -91,7 +94,7 @@ char *const *parse_plugin(const char *plugin, const char *host, const char *serv
 	unsigned int service_len = strlen(service);
 
 	for (c = p = plugin; *c; c++)
-	{	if (isspace(*c) && !isspace(*p))
+	{	if (isspace((unsigned char)*c) && !isspace((unsigned char)*p))
 			s += sizeof(char*);
 		if (*p == '%' && *c == 'h')
 			host_count++;
@@ -101,7 +104,7 @@ char *const *parse_plugin(const char *plugin, const char *host, const char *serv
 	}
 
 	plugin_copy_len = plugin_len + host_len * host_count + service_len * service_count;
-	plugin_copy = malloc(plugin_copy_len + 1);
+	plugin_copy = (char *)malloc(plugin_copy_len + 1);
 	if (!plugin_copy)
 	{
 		report(stderr, GT_("fetchmail: malloc failed\n"));
@@ -127,25 +130,25 @@ char *const *parse_plugin(const char *plugin, const char *host, const char *serv
 	}
 	plugin_copy[plugin_copy_len] = 0;
 
-	argvec = malloc(s);
+	argvec = (char **)malloc(s);
 	if (!argvec)
 	{
 		report(stderr, GT_("fetchmail: malloc failed\n"));
 		return NULL;
 	}
 	memset(argvec, 0, s);
-	for (c = p = plugin_copy, i = 0; *c; c++)
-	{	if ((!isspace(*c)) && (c == p ? 1 : isspace(*p))) {
-			argvec[i] = c;
+	for (p = cp = plugin_copy, i = 0; *cp; cp++)
+	{	if ((!isspace((unsigned char)*cp)) && (cp == p ? 1 : isspace((unsigned char)*p))) {
+			argvec[i] = cp;
 			i++;
 		}
-		p = c;
+		p = cp;
 	}
 	for (cp = plugin_copy; *cp; cp++)
-	{	if (isspace(*cp))
+	{	if (isspace((unsigned char)*cp))
 			*cp = 0;
 	}
-	return (char *const*)argvec;
+	return argvec;
 }
 
 static int handle_plugin(const char *host,
@@ -170,14 +173,13 @@ static int handle_plugin(const char *host,
 		/* error */
 		report(stderr, GT_("fetchmail: fork failed\n"));
 		return -1;
-		break;
 	case 0:	/* child */
 		/* fds[1] is the parent's end; close it for proper EOF
 		** detection */
 		(void) close(fds[1]);
 		if ( (dup2(fds[0],0) == -1) || (dup2(fds[0],1) == -1) ) {
 			report(stderr, GT_("dup2 failed\n"));
-			exit(1);
+			_exit(EXIT_FAILURE);
 		}
 		/* fds[0] is now connected to 0 and 1; close it */
 		(void) close(fds[0]);
@@ -186,7 +188,7 @@ static int handle_plugin(const char *host,
 		argvec = parse_plugin(plugin,host,service);
 		execvp(*argvec, argvec);
 		report(stderr, GT_("execvp(%s) failed\n"), *argvec);
-		exit(0);
+		_exit(EXIT_FAILURE);
 		break;
 	default:	/* parent */
 		/* NOP */
@@ -198,30 +200,11 @@ static int handle_plugin(const char *host,
 }
 #endif /* HAVE_SOCKETPAIR */
 
-#ifdef __UNUSED__
-#include <sys/time.h>
-
-int SockCheckOpen(int fd)
-/* poll given socket; is it selectable? */
-{
-    fd_set r, w, e;
-    int rt;
-    struct timeval tv;
-  
-    for (;;) 
-    {
-	FD_ZERO(&r); FD_ZERO(&w); FD_ZERO(&e);
-	FD_SET(fd, &e);
-    
-	tv.tv_sec = 0; tv.tv_usec = 0;
-	rt = select(fd+1, &r, &w, &e, &tv);
-	if (rt == -1 && (errno != EAGAIN && errno != EINTR))
-	    return 0;
-	if (rt != -1)
-	    return 1;
-    }
+/** Set socket to SO_KEEPALIVE. \return 0 for success. */
+int SockKeepalive(int sock) {
+    int keepalive = 1;
+    return setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, &keepalive, sizeof keepalive);
 }
-#endif /* __UNUSED__ */
 
 int UnixOpen(const char *path)
 {
@@ -238,12 +221,12 @@ int UnixOpen(const char *path)
 	return -1;
     }
 
-	/* Socket opened saved. Usefull if connect timeout 
-	 * because it can be closed.
-	 */
-	mailserver_socket_temp = sock;
-    
-	if (connect(sock, (struct sockaddr *) &ad, sizeof(ad)) < 0)
+    /* Socket opened saved. Usefull if connect timeout 
+     * because it can be closed.
+     */
+    mailserver_socket_temp = sock;
+
+    if (connect(sock, (struct sockaddr *) &ad, sizeof(ad)) < 0)
     {
 	int olderr = errno;
 	fm_close(sock);	/* don't use SockClose, no traffic yet */
@@ -251,225 +234,114 @@ int UnixOpen(const char *path)
 	errno = olderr;
 	sock = -1;
     }
-	
-	/* No connect timeout, then no need to set mailserver_socket_temp */
-	mailserver_socket_temp = -1;
+
+    /* No connect timeout, then no need to set mailserver_socket_temp */
+    mailserver_socket_temp = -1;
 
     return sock;
 }
 
-#if INET6_ENABLE
-int SockOpen(const char *host, const char *service, const char *options,
-	     const char *plugin)
+int SockOpen(const char *host, const char *service,
+	     const char *plugin, struct addrinfo **ai0)
 {
-    struct addrinfo *ai, *ai0, req;
-    int i;
-#if NET_SECURITY
-    void *request = NULL;
-    int requestlen;
-#endif /* NET_SECURITY */
+    struct addrinfo *ai, req;
+    int i, acterr = 0;
+    int ord;
+    char errbuf[8192] = "";
 
 #ifdef HAVE_SOCKETPAIR
     if (plugin)
 	return handle_plugin(host,service,plugin);
 #endif /* HAVE_SOCKETPAIR */
+
     memset(&req, 0, sizeof(struct addrinfo));
     req.ai_socktype = SOCK_STREAM;
+#ifdef AI_ADDRCONFIG
+    req.ai_flags = AI_ADDRCONFIG;
+#endif
 
-    if (getaddrinfo(host, service, &req, &ai0)) {
-	report(stderr, GT_("fetchmail: getaddrinfo(%s.%s)\n"), host,service);
+    i = fm_getaddrinfo(host, service, &req, ai0);
+    if (i) {
+	report(stderr, GT_("getaddrinfo(\"%s\",\"%s\") error: %s\n"),
+		host, service, gai_strerror(i));
+	if (i == EAI_SERVICE)
+	    report(stderr, GT_("Try adding the --service option (see also FAQ item R12).\n"));
 	return -1;
     }
 
-#if NET_SECURITY
-    if (!options)
-	requestlen = 0;
-    else
-	if (net_security_strtorequest((char *)options, &request, &requestlen))
-	    goto ret;
-
-    i = inner_connect(ai0, request, requestlen, NULL, NULL, "fetchmail", NULL);
-    if (request)
-	free(request);
-
- ret:
-#else /* NET_SECURITY */
-#ifdef HAVE_INNER_CONNECT
-    i = inner_connect(ai0, NULL, 0, NULL, NULL, "fetchmail", NULL);
-    if (i >= 0)
-	break;
-#else
-
+    /* NOTE a Linux bug here - getaddrinfo will happily return 127.0.0.1
+     * twice if no IPv6 is configured */
     i = -1;
-    for (ai = ai0; ai; ai = ai->ai_next) {
-	i = socket(ai->ai_family, ai->ai_socktype, 0);
-	if (i < 0)
-	    continue;
+    for (ord = 0, ai = *ai0; ai; ord++, ai = ai->ai_next) {
+	char buf[256]; /* hostname */
+	char pb[256];  /* service name */
+	int gnie;      /* getnameinfo result code */
 
-	/* Socket opened saved. Usefull if connect timeout 
-	 * because it can be closed.
-	 */
+	gnie = getnameinfo(ai->ai_addr, ai->ai_addrlen, buf, sizeof(buf), NULL, 0, NI_NUMERICHOST);
+	if (gnie)
+	    snprintf(buf, sizeof(buf), GT_("unknown (%s)"), gai_strerror(gnie));
+	gnie = getnameinfo(ai->ai_addr, ai->ai_addrlen, NULL, 0, pb, sizeof(pb), NI_NUMERICSERV);
+	if (gnie)
+	    snprintf(pb, sizeof(pb), GT_("unknown (%s)"), gai_strerror(gnie));
+
+	if (outlevel >= O_VERBOSE)
+	    report_build(stdout, GT_("Trying to connect to %s/%s..."), buf, pb);
+	i = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
+	if (i < 0) {
+	    int e = errno;
+	    /* mask EAFNOSUPPORT errors, they confuse users for
+	     * multihomed hosts */
+	    if (errno != EAFNOSUPPORT)
+		acterr = errno;
+	    if (outlevel >= O_VERBOSE)
+		report_complete(stdout, GT_("cannot create socket: %s\n"), strerror(e));
+	    snprintf(errbuf+strlen(errbuf), sizeof(errbuf)-strlen(errbuf),\
+		     GT_("name %d: cannot create socket family %d type %d: %s\n"), ord, ai->ai_family, ai->ai_socktype, strerror(e));
+	    continue;
+	}
+
+	SockKeepalive(i);
+
+	/* Save socket descriptor.
+	 * Used to close the socket after connect timeout. */
 	mailserver_socket_temp = i;
 
 	if (connect(i, (struct sockaddr *) ai->ai_addr, ai->ai_addrlen) < 0) {
+	    int e = errno;
+
+	    /* additionally, suppress IPv4 network unreach errors */
+	    if (e != EAFNOSUPPORT)
+		acterr = errno;
+
+	    if (outlevel >= O_VERBOSE)
+		report_complete(stdout, GT_("connection failed.\n"));
+	    if (outlevel >= O_VERBOSE)
+		report(stderr, GT_("connection to %s:%s [%s/%s] failed: %s.\n"), host, service, buf, pb, strerror(e));
+	    snprintf(errbuf+strlen(errbuf), sizeof(errbuf)-strlen(errbuf), GT_("name %d: connection to %s:%s [%s/%s] failed: %s.\n"), ord, host, service, buf, pb, strerror(e));
 	    fm_close(i);
 	    i = -1;
 	    continue;
+	} else {
+	    if (outlevel >= O_VERBOSE)
+		report_complete(stdout, GT_("connected.\n"));
 	}
-	
+
 	/* No connect timeout, then no need to set mailserver_socket_temp */
 	mailserver_socket_temp = -1;
-	
+
 	break;
     }
 
-#endif
-#endif /* NET_SECURITY */
+    fm_freeaddrinfo(*ai0);
+    *ai0 = NULL;
 
-    freeaddrinfo(ai0);
+    if (i == -1) {
+	report(stderr, GT_("Connection errors for this poll:\n%s"), errbuf);
+	errno = acterr;
+    }
 
     return i;
 }
-#else /* INET6_ENABLE */
-#ifndef HAVE_INET_ATON
-#ifndef  INADDR_NONE
-#ifdef   INADDR_BROADCAST
-#define  INADDR_NONE	INADDR_BROADCAST
-#else
-#define	 INADDR_NONE	-1
-#endif
-#endif
-#endif /* HAVE_INET_ATON */
-
-int SockOpen(const char *host, int clientPort, const char *options,
-	     const char *plugin)
-{
-    int sock = -1;	/* pacify -Wall */
-#ifndef HAVE_INET_ATON
-    unsigned long inaddr;
-#endif /* HAVE_INET_ATON */
-    struct sockaddr_in ad, **pptr;
-    struct hostent *hp;
-
-#ifdef HAVE_SOCKETPAIR
-    if (plugin) {
-      char buf[10];
-#ifdef HAVE_SNPRINTF
-      snprintf(buf, sizeof(buf),  /* Yeah, paranoic. So what? :P */
-#else
-      sprintf(buf,
-#endif /* HAVE_SNPRINTF */
-	      "%d",clientPort);
-      return handle_plugin(host,buf,plugin);
-    }
-#endif /* HAVE_SOCKETPAIR */
-
-    memset(&ad, 0, sizeof(ad));
-    ad.sin_family = AF_INET;
-
-    /* we'll accept a quad address */
-#ifndef HAVE_INET_ATON
-    inaddr = inet_addr((char*)host);
-    if (inaddr != INADDR_NONE)
-    {
-        memcpy(&ad.sin_addr, &inaddr, sizeof(inaddr));
-#else
-    if (inet_aton(host, &ad.sin_addr))
-    {
-#endif /* HAVE_INET_ATON */
-        ad.sin_port = htons(clientPort);
-
-        sock = socket(AF_INET, SOCK_STREAM, 0);
-        if (sock < 0)
-        {
-            h_errno = 0;
-            return -1;
-        }
-
-		/* Socket opened saved. Usefull if connect timeout because
-		 * it can be closed
-		 */
-		mailserver_socket_temp = sock;
-		
-        if (connect(sock, (struct sockaddr *) &ad, sizeof(ad)) < 0)
-        {
-            int olderr = errno;
-            fm_close(sock);	/* don't use SockClose, no traffic yet */
-            h_errno = 0;
-            errno = olderr;
-            return -1;
-        }
-
-		/* No connect timeout, then no need to set mailserver_socket_temp */
-		mailserver_socket_temp = -1;
-		
-#ifndef HAVE_INET_ATON
-    }
-#else
-    }
-#endif /* HAVE_INET_ATON */
-    else {
-        hp = gethostbyname((char*)host);
-
-        if (hp == NULL)
-	{
-	    errno = 0;
-	    return -1;
-	}
-	/*
-	 * Add a check to make sure the address has a valid IPv4 or IPv6
-	 * length.  This prevents buffer spamming by a broken DNS.
-	 */
-	if(hp->h_length != 4 && hp->h_length != 8)
-	{
-	    h_errno = errno = 0;
-	    report(stderr, 
-		   GT_("fetchmail: illegal address length received for host %s\n"),host);
-	    return -1;
-	}
-	/*
-	 * Try all addresses of a possibly multihomed host until we get
-	 * a successful connect or until we run out of addresses.
-	 */
-	pptr = (struct sockaddr_in **)hp->h_addr_list;
-	for(; *pptr != NULL; pptr++)
-	{
-	    sock = socket(AF_INET, SOCK_STREAM, 0);
-	    if (sock < 0)
-	    {
-		h_errno = 0;
-		return -1;
-	    }
-
-		/* Socket opened saved. Usefull if connect timeout because
-		 * it can be closed
-		 */
-		mailserver_socket_temp = sock;
-		
-	    ad.sin_port = htons(clientPort);
-	    memcpy(&ad.sin_addr, *pptr, sizeof(struct in_addr));
-	    if (connect(sock, (struct sockaddr *) &ad, sizeof(ad)) == 0) {
-			/* No connect timeout, then no need to set mailserver_socket_temp */
-			mailserver_socket_temp = -1;
-			break; /* success */
-		}	
-	    fm_close(sock);	/* don't use SockClose, no traffic yet */
-	    memset(&ad, 0, sizeof(ad));
-	    ad.sin_family = AF_INET;
-	}
-	if(*pptr == NULL)
-	{
-	    int olderr = errno;
-	    fm_close(sock);	/* don't use SockClose, no traffic yet */
-	    h_errno = 0;
-	    errno = olderr;
-	    return -1;
-	}
-    }
-
-    return(sock);
-}
-#endif /* INET6_ENABLE */
 
 
 #if defined(HAVE_STDARG_H)
@@ -490,29 +362,26 @@ va_dcl {
 #else
     va_start(ap);
 #endif
-#ifdef HAVE_VSNPRINTF
     vsnprintf(buf, sizeof(buf), format, ap);
-#else
-    vsprintf(buf, format, ap);
-#endif
     va_end(ap);
     return SockWrite(sock, buf, strlen(buf));
 
 }
 
 #ifdef SSL_ENABLE
-#include "openssl/ssl.h"
-#include "openssl/err.h"
-#include "openssl/pem.h"
-#include "openssl/x509.h"
+#include <openssl/ssl.h>
+#include <openssl/err.h>
+#include <openssl/pem.h>
+#include <openssl/x509v3.h>
+#include <openssl/rand.h>
 
-static	SSL_CTX *_ctx = NULL;
+static	SSL_CTX *_ctx[FD_SETSIZE];
 static	SSL *_ssl_context[FD_SETSIZE];
 
-SSL	*SSLGetContext( int );
+static SSL	*SSLGetContext( int );
 #endif /* SSL_ENABLE */
 
-int SockWrite(int sock, char *buf, int len)
+int SockWrite(int sock, const char *buf, int len)
 {
     int n, wrlen = 0;
 #ifdef	SSL_ENABLE
@@ -567,7 +436,7 @@ int SockRead(int sock, char *buf, int len)
 		/* OK...  SSL_peek works a little different from MSG_PEEK
 			Problem is that SSL_peek can return 0 if there
 			is no data currently available.  If, on the other
-			hand, we loose the socket, we also get a zero, but
+			hand, we lose the socket, we also get a zero, but
 			the SSL_read then SEGFAULTS!  To deal with this,
 			we'll check the error code any time we get a return
 			of zero from SSL_peek.  If we have an error, we bail.
@@ -583,7 +452,7 @@ int SockRead(int sock, char *buf, int len)
 			/* SSL_peek says no data...  Does he mean no data
 			or did the connection blow up?  If we got an error
 			then bail! */
-			if( 0 != ( n = SSL_get_error(ssl, n) ) ) {
+			if (0 != SSL_get_error(ssl, n)) {
 				return -1;
 			}
 			/* We didn't get an error so read at least one
@@ -593,10 +462,10 @@ int SockRead(int sock, char *buf, int len)
 			 * We don't have a string to pass through
 			 * the strchr at this point yet */
 			newline = NULL;
-		} else if ((newline = memchr(bp, '\n', n)) != NULL)
+		} else if ((newline = (char *)memchr(bp, '\n', n)) != NULL)
 			n = newline - bp + 1;
 		/* Matthias Andree: SSL_read can return 0, in that case
-		 * we must cal SSL_get_error to figure if there was
+		 * we must call SSL_get_error to figure if there was
 		 * an error or just a "no data" condition */
 		if ((n = SSL_read(ssl, bp, n)) <= 0) {
 			if ((n = SSL_get_error(ssl, n))) {
@@ -622,7 +491,7 @@ int SockRead(int sock, char *buf, int len)
 	    if ((n = fm_peek(sock, bp, len)) <= 0)
 #endif
 		return (-1);
-	    if ((newline = memchr(bp, '\n', n)) != NULL)
+	    if ((newline = (char *)memchr(bp, '\n', n)) != NULL)
 		n = newline - bp + 1;
 #ifndef __BEOS__
 	    if ((n = fm_read(sock, bp, n)) == -1)
@@ -634,6 +503,7 @@ int SockRead(int sock, char *buf, int len)
     } while 
 	    (!newline && len);
     *bp = '\0';
+
     return bp - buf;
 }
 
@@ -662,7 +532,7 @@ int SockPeek(int sock)
 			/* SSL_peek says 0...  Does that mean no data
 			or did the connection blow up?  If we got an error
 			then bail! */
-			if( 0 != ( n = SSL_get_error(ssl, n) ) ) {
+			if(0 != SSL_get_error(ssl, n)) {
 				return -1;
 			}
 
@@ -694,29 +564,34 @@ static	int _check_fp;
 static	char *_check_digest;
 static 	char *_server_label;
 static	int _depth0ck;
+static	int _firstrun;
+static	int _prev_err;
+static	int _verify_ok;
 
 SSL *SSLGetContext( int sock )
 {
-	/* If SSLOpen has never initialized - just return NULL */
-	if( NULL == _ctx )
+	if( sock < 0 || (unsigned)sock > FD_SETSIZE )
 		return NULL;
-
-	if( sock < 0 || sock > FD_SETSIZE )
+	if( _ctx[sock] == NULL )
 		return NULL;
 	return _ssl_context[sock];
 }
 
-
-int SSL_verify_callback( int ok_return, X509_STORE_CTX *ctx, int strict )
+/* ok_return (preverify_ok) is 1 if this stage of certificate verification
+   passed, or 0 if it failed. This callback lets us display informative
+   errors, and perform additional validation (e.g. CN matches) */
+static int SSL_verify_callback( int ok_return, X509_STORE_CTX *ctx, int strict )
 {
+#define SSLverbose (((outlevel) >= O_DEBUG) || ((outlevel) >= O_VERBOSE && (depth) == 0)) 
 	char buf[257];
 	X509 *x509_cert;
-	int err, depth;
+	int err, depth, i;
 	unsigned char digest[EVP_MAX_MD_SIZE];
 	char text[EVP_MAX_MD_SIZE * 3 + 1], *tp, *te;
-	EVP_MD *digest_tp;
-	unsigned int dsz, i, esz;
+	const EVP_MD *digest_tp;
+	unsigned int dsz, esz;
 	X509_NAME *subj, *issuer;
+	char *tt;
 
 	x509_cert = X509_STORE_CTX_get_current_cert(ctx);
 	err = X509_STORE_CTX_get_error(ctx);
@@ -725,56 +600,115 @@ int SSL_verify_callback( int ok_return, X509_STORE_CTX *ctx, int strict )
 	subj = X509_get_subject_name(x509_cert);
 	issuer = X509_get_issuer_name(x509_cert);
 
-	if (depth == 0) {
-		_depth0ck = 1;
-		
-		if (outlevel == O_VERBOSE) {
+	if (outlevel >= O_VERBOSE) {
+		if (depth == 0 && SSLverbose)
+			report(stdout, GT_("Server certificate:\n"));
+		else {
+			if (_firstrun) {
+				_firstrun = 0;
+				if (SSLverbose)
+					report(stdout, GT_("Certificate chain, from root to peer, starting at depth %d:\n"), depth);
+			} else {
+				if (SSLverbose)
+					report(stdout, GT_("Certificate at depth %d:\n"), depth);
+			}
+		}
+
+		if (SSLverbose) {
 			if ((i = X509_NAME_get_text_by_NID(issuer, NID_organizationName, buf, sizeof(buf))) != -1) {
-				report(stdout, GT_("Issuer Organization: %s\n"), buf);
-				if (i >= sizeof(buf) - 1)
+				report(stdout, GT_("Issuer Organization: %s\n"), (tt = sdump(buf, i)));
+				xfree(tt);
+				if ((size_t)i >= sizeof(buf) - 1)
 					report(stdout, GT_("Warning: Issuer Organization Name too long (possibly truncated).\n"));
 			} else
 				report(stdout, GT_("Unknown Organization\n"));
 			if ((i = X509_NAME_get_text_by_NID(issuer, NID_commonName, buf, sizeof(buf))) != -1) {
-				report(stdout, GT_("Issuer CommonName: %s\n"), buf);
-				if (i >= sizeof(buf) - 1)
+				report(stdout, GT_("Issuer CommonName: %s\n"), (tt = sdump(buf, i)));
+				xfree(tt);
+				if ((size_t)i >= sizeof(buf) - 1)
 					report(stdout, GT_("Warning: Issuer CommonName too long (possibly truncated).\n"));
 			} else
 				report(stdout, GT_("Unknown Issuer CommonName\n"));
 		}
+	}
+
+	if ((i = X509_NAME_get_text_by_NID(subj, NID_commonName, buf, sizeof(buf))) != -1) {
+		if (SSLverbose) {
+			report(stdout, GT_("Subject CommonName: %s\n"), (tt = sdump(buf, i)));
+			xfree(tt);
+		}
+		if ((size_t)i >= sizeof(buf) - 1) {
+			/* Possible truncation. In this case, this is a DNS name, so this
+			 * is really bad. We do not tolerate this even in the non-strict case. */
+			report(stderr, GT_("Bad certificate: Subject CommonName too long!\n"));
+			return (0);
+		}
+		if ((size_t)i > strlen(buf)) {
+			/* Name contains embedded NUL characters, so we complain. This is likely
+			 * a certificate spoofing attack. */
+			report(stderr, GT_("Bad certificate: Subject CommonName contains NUL, aborting!\n"));
+			return 0;
+		}
+	}
+
+	if (depth == 0) { /* peer certificate */
+		if (!_depth0ck) {
+			_depth0ck = 1;
+		}
+
 		if ((i = X509_NAME_get_text_by_NID(subj, NID_commonName, buf, sizeof(buf))) != -1) {
-			if (outlevel == O_VERBOSE)
-				report(stdout, GT_("Server CommonName: %s\n"), buf);
-			if (i >= sizeof(buf) - 1) {
-				/* Possible truncation. In this case, this is a DNS name, so this
-				 * is really bad. We do not tolerate this even in the non-strict case. */
-				report(stderr, GT_("Bad certificate: Subject CommonName too long!\n"));
-				return (0);
-			}
 			if (_ssl_server_cname != NULL) {
 				char *p1 = buf;
 				char *p2 = _ssl_server_cname;
-				int n;
-				
-				if (*p1 == '*') {
-					++p1;
-					n = strlen(p2) - strlen(p1);
-					if (n >= 0)
-						p2 += n;
-				}	
-				if (0 != strcasecmp(p1, p2)) {
-					report(stderr,
-					    GT_("Server CommonName mismatch: %s != %s\n"),
-					    buf, _ssl_server_cname );
-					if (ok_return && strict)
-						return (0);
+				int matched = 0;
+				STACK_OF(GENERAL_NAME) *gens;
+
+				/* RFC 2595 section 2.4: find a matching name
+				 * first find a match among alternative names */
+				gens = (STACK_OF(GENERAL_NAME) *)X509_get_ext_d2i(x509_cert, NID_subject_alt_name, NULL, NULL);
+				if (gens) {
+					int j, r;
+					for (j = 0, r = sk_GENERAL_NAME_num(gens); j < r; ++j) {
+						const GENERAL_NAME *gn = sk_GENERAL_NAME_value(gens, j);
+						if (gn->type == GEN_DNS) {
+							char *pp1 = (char *)gn->d.ia5->data;
+							char *pp2 = _ssl_server_cname;
+							if (outlevel >= O_VERBOSE) {
+								report(stdout, GT_("Subject Alternative Name: %s\n"), (tt = sdump(pp1, (size_t)gn->d.ia5->length)));
+								xfree(tt);
+							}
+							/* Name contains embedded NUL characters, so we complain. This
+							 * is likely a certificate spoofing attack. */
+							if ((size_t)gn->d.ia5->length != strlen(pp1)) {
+								report(stderr, GT_("Bad certificate: Subject Alternative Name contains NUL, aborting!\n"));
+								sk_GENERAL_NAME_free(gens);
+								return 0;
+							}
+							if (name_match(pp1, pp2)) {
+							    matched = 1;
+							}
+						}
+					}
+					sk_GENERAL_NAME_free(gens);
 				}
-			} else if (ok_return && strict) {
+				if (name_match(p1, p2)) {
+					matched = 1;
+				}
+				if (!matched) {
+					if (strict || SSLverbose) {
+						report(stderr,
+								GT_("Server CommonName mismatch: %s != %s\n"),
+								(tt = sdump(buf, i)), _ssl_server_cname );
+						xfree(tt);
+					}
+					ok_return = 0;
+				}
+			} else if (ok_return) {
 				report(stderr, GT_("Server name not set, could not verify certificate!\n"));
-				return (0);
+				if (strict) return (0);
 			}
 		} else {
-			if (outlevel == O_VERBOSE)
+			if (outlevel >= O_VERBOSE)
 				report(stdout, GT_("Unknown Server CommonName\n"));
 			if (ok_return && strict) {
 				report(stderr, GT_("Server name not specified in certificate!\n"));
@@ -783,8 +717,10 @@ int SSL_verify_callback( int ok_return, X509_STORE_CTX *ctx, int strict )
 		}
 		/* Print the finger print. Note that on errors, we might print it more than once
 		 * normally; we kluge around that by using a global variable. */
-		if (_check_fp) {
-			_check_fp = 0;
+		if (_check_fp == 1) {
+			unsigned dp;
+
+			_check_fp = -1;
 			digest_tp = EVP_md5();
 			if (digest_tp == NULL) {
 				report(stderr, GT_("EVP_md5() failed!\n"));
@@ -796,13 +732,9 @@ int SSL_verify_callback( int ok_return, X509_STORE_CTX *ctx, int strict )
 			}
 			tp = text;
 			te = text + sizeof(text);
-			for (i = 0; i < dsz; i++) {
-#ifdef HAVE_SNPRINTF
-				esz = snprintf(tp, te - tp, i > 0 ? ":%02X" : "%02X", digest[i]);
-#else
-				esz = sprintf(tp, i > 0 ? ":%02X" : "%02X", digest[i]);
-#endif
-				if (esz >= te - tp) {
+			for (dp = 0; dp < dsz; dp++) {
+				esz = snprintf(tp, te - tp, dp > 0 ? ":%02X" : "%02X", digest[dp]);
+				if (esz >= (size_t)(te - tp)) {
 					report(stderr, GT_("Digest text buffer too small!\n"));
 					return (0);
 				}
@@ -811,101 +743,203 @@ int SSL_verify_callback( int ok_return, X509_STORE_CTX *ctx, int strict )
 			if (outlevel > O_NORMAL)
 			    report(stdout, GT_("%s key fingerprint: %s\n"), _server_label, text);
 			if (_check_digest != NULL) {
-				if (strcmp(text, _check_digest) == 0) {
+				if (strcasecmp(text, _check_digest) == 0) {
 				    if (outlevel > O_NORMAL)
 					report(stdout, GT_("%s fingerprints match.\n"), _server_label);
 				} else {
-				    if (outlevel > O_SILENT)
-					report(stderr, GT_("%s fingerprints do not match!\n"), _server_label);
+				    report(stderr, GT_("%s fingerprints do not match!\n"), _server_label);
 				    return (0);
 				}
-			}
-		}
-	}
+			} /* if (_check_digest != NULL) */
+		} /* if (_check_fp) */
+	} /* if (depth == 0 && !_depth0ck) */
+
+	if (err != X509_V_OK && err != _prev_err && !(_check_fp != 0 && _check_digest && !strict)) {
+		_prev_err = err;
+					
+                report(stderr, GT_("Server certificate verification error: %s\n"), X509_verify_cert_error_string(err));
+                /* We gave the error code, but maybe we can add some more details for debugging */
 
-	if (err != X509_V_OK && (strict || outlevel == O_VERBOSE)) {
-		report(strict ? stderr : stdout, GT_("Warning: server certificate verification: %s\n"), X509_verify_cert_error_string(err));
-		/* We gave the error code, but maybe we can add some more details for debugging */
 		switch (err) {
 		case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:
 			X509_NAME_oneline(issuer, buf, sizeof(buf));
 			buf[sizeof(buf) - 1] = '\0';
-			report(stdout, GT_("unknown issuer (first %d characters): %s\n"), sizeof(buf), buf);
+			report(stderr, GT_("unknown issuer (first %d characters): %s\n"), (int)(sizeof(buf)-1), buf);
+			report(stderr, GT_("This error usually happens when the server provides an incomplete certificate "
+						"chain, which is nothing fetchmail could do anything about.  For details, "
+						"please see the README.SSL-SERVER document that comes with fetchmail.\n"));
+			break;
+		case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
+		case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
+		case X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN:
+			X509_NAME_oneline(subj, buf, sizeof(buf));
+			buf[sizeof(buf) - 1] = '\0';
+			report(stderr, GT_("This means that the root signing certificate (issued for %s) is not in the "
+						"trusted CA certificate locations, or that c_rehash needs to be run "
+						"on the certificate directory. For details, please "
+						"see the documentation of --sslcertpath and --sslcertfile in the manual page.\n"), buf);
+			break;
+		default:
 			break;
 		}
 	}
+	/*
+	 * If not in strict checking mode (--sslcertck), override this
+	 * and pretend that verification had succeeded.
+	 */
+	_verify_ok &= ok_return;
 	if (!strict)
 		ok_return = 1;
 	return (ok_return);
 }
 
-int SSL_nock_verify_callback( int ok_return, X509_STORE_CTX *ctx )
+static int SSL_nock_verify_callback( int ok_return, X509_STORE_CTX *ctx )
 {
 	return SSL_verify_callback(ok_return, ctx, 0);
 }
 
-int SSL_ck_verify_callback( int ok_return, X509_STORE_CTX *ctx )
+static int SSL_ck_verify_callback( int ok_return, X509_STORE_CTX *ctx )
 {
 	return SSL_verify_callback(ok_return, ctx, 1);
 }
 
+
+/* get commonName from certificate set in file.
+ * commonName is stored in buffer namebuffer, limited with namebufferlen
+ */
+static const char *SSLCertGetCN(const char *mycert,
+                                char *namebuffer, size_t namebufferlen)
+{
+	const char *ret       = NULL;
+	BIO        *certBio   = NULL;
+	X509       *x509_cert = NULL;
+	X509_NAME  *certname  = NULL;
+
+	if (namebuffer && namebufferlen > 0) {
+		namebuffer[0] = 0x00;
+		certBio = BIO_new_file(mycert,"r");
+		if (certBio) {
+			x509_cert = PEM_read_bio_X509(certBio,NULL,NULL,NULL);
+			BIO_free(certBio);
+		}
+		if (x509_cert) {
+			certname = X509_get_subject_name(x509_cert);
+			if (certname &&
+			    X509_NAME_get_text_by_NID(certname, NID_commonName,
+						      namebuffer, namebufferlen) > 0)
+				ret = namebuffer;
+			X509_free(x509_cert);
+		}
+	}
+	return ret;
+}
+
 /* performs initial SSL handshake over the connected socket
  * uses SSL *ssl global variable, which is currently defined
  * in this file
  */
-int SSLOpen(int sock, char *mycert, char *mykey, char *myproto, int certck, char *certpath,
-    char *fingerprint, char *servercname, char *label)
+int SSLOpen(int sock, char *mycert, char *mykey, const char *myproto, int certck,
+    char *cacertfile, char *certpath,
+    char *fingerprint, char *servercname, char *label, char **remotename)
 {
-	SSL *ssl;
-	
+        struct stat randstat;
+        int i;
+
 	SSL_load_error_strings();
-	SSLeay_add_ssl_algorithms();
-	
-	if( sock < 0 || sock > FD_SETSIZE ) {
+	SSL_library_init();
+	OpenSSL_add_all_algorithms(); /* see Debian Bug#576430 and manpage */
+
+        if (stat("/dev/random", &randstat)  &&
+            stat("/dev/urandom", &randstat)) {
+          /* Neither /dev/random nor /dev/urandom are present, so add
+             entropy to the SSL PRNG a hard way. */
+          for (i = 0; i < 10000  &&  ! RAND_status (); ++i) {
+            char buf[4];
+            struct timeval tv;
+            gettimeofday (&tv, 0);
+            buf[0] = tv.tv_usec & 0xF;
+            buf[2] = (tv.tv_usec & 0xF0) >> 4;
+            buf[3] = (tv.tv_usec & 0xF00) >> 8;
+            buf[1] = (tv.tv_usec & 0xF000) >> 12;
+            RAND_add (buf, sizeof buf, 0.1);
+          }
+        }
+
+	if( sock < 0 || (unsigned)sock > FD_SETSIZE ) {
 		report(stderr, GT_("File descriptor out of range for SSL") );
 		return( -1 );
 	}
 
-	if( ! _ctx ) {
-		/* Be picky and make sure the memory is cleared */
-		memset( _ssl_context, 0, sizeof( _ssl_context ) );
-		if(myproto) {
-			if(!strcmp("ssl2",myproto)) {
-				_ctx = SSL_CTX_new(SSLv2_client_method());
-			} else if(!strcmp("ssl3",myproto)) {
-				_ctx = SSL_CTX_new(SSLv3_client_method());
-			} else if(!strcmp("tls1",myproto)) {
-				_ctx = SSL_CTX_new(TLSv1_client_method());
-			} else if (!strcmp("ssl23",myproto)) {
-				myproto = NULL;
-			} else {
-				fprintf(stderr,GT_("Invalid SSL protocol '%s' specified, using default (SSLv23).\n"), myproto);
-				myproto = NULL;
-			}
-		}
-		if(!myproto) {
-			_ctx = SSL_CTX_new(SSLv23_client_method());
-		}
-		if(_ctx == NULL) {
-			ERR_print_errors_fp(stderr);
-			return(-1);
+	/* Make sure a connection referring to an older context is not left */
+	_ssl_context[sock] = NULL;
+	if(myproto) {
+		if(!strcasecmp("ssl2",myproto)) {
+#if HAVE_DECL_SSLV2_CLIENT_METHOD + 0 > 0
+			_ctx[sock] = SSL_CTX_new(SSLv2_client_method());
+#else
+			report(stderr, GT_("Your operating system does not support SSLv2.\n"));
+			return -1;
+#endif
+		} else if(!strcasecmp("ssl3",myproto)) {
+			_ctx[sock] = SSL_CTX_new(SSLv3_client_method());
+		} else if(!strcasecmp("tls1",myproto)) {
+			_ctx[sock] = SSL_CTX_new(TLSv1_client_method());
+		} else if (!strcasecmp("ssl23",myproto)) {
+			myproto = NULL;
+		} else {
+			fprintf(stderr,GT_("Invalid SSL protocol '%s' specified, using default (SSLv23).\n"), myproto);
+			myproto = NULL;
 		}
 	}
+	if(!myproto) {
+		_ctx[sock] = SSL_CTX_new(SSLv23_client_method());
+	}
+	if(_ctx[sock] == NULL) {
+		ERR_print_errors_fp(stderr);
+		return(-1);
+	}
+
+	SSL_CTX_set_options(_ctx[sock], SSL_OP_ALL);
+
+	{
+	    char *tmp = getenv("FETCHMAIL_DISABLE_CBC_IV_COUNTERMEASURE");
+	    if (tmp == NULL || *tmp == '\0' || strspn(tmp, " \t") == strlen(tmp))
+		SSL_CTX_clear_options(_ctx[sock], SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS);
+	}
 
 	if (certck) {
-		SSL_CTX_set_verify(_ctx, SSL_VERIFY_PEER, SSL_ck_verify_callback);
-		if (certpath)
-			SSL_CTX_load_verify_locations(_ctx, NULL, certpath);
+		SSL_CTX_set_verify(_ctx[sock], SSL_VERIFY_PEER, SSL_ck_verify_callback);
 	} else {
 		/* In this case, we do not fail if verification fails. However,
-		 *  we provide the callback for output and possible fingerprint checks. */
-		SSL_CTX_set_verify(_ctx, SSL_VERIFY_PEER, SSL_nock_verify_callback);
+		 * we provide the callback for output and possible fingerprint
+		 * checks. */
+		SSL_CTX_set_verify(_ctx[sock], SSL_VERIFY_PEER, SSL_nock_verify_callback);
+	}
+
+	/* Check which trusted X.509 CA certificate store(s) to load */
+	{
+		char *tmp;
+		int want_default_cacerts = 0;
+
+		/* Load user locations if any is given */
+		if (certpath || cacertfile)
+			SSL_CTX_load_verify_locations(_ctx[sock],
+						cacertfile, certpath);
+		else
+			want_default_cacerts = 1;
+
+		tmp = getenv("FETCHMAIL_INCLUDE_DEFAULT_X509_CA_CERTS");
+		if (want_default_cacerts || (tmp && tmp[0])) {
+			SSL_CTX_set_default_verify_paths(_ctx[sock]);
+		}
 	}
 	
-	_ssl_context[sock] = SSL_new(_ctx);
+	_ssl_context[sock] = SSL_new(_ctx[sock]);
 	
 	if(_ssl_context[sock] == NULL) {
 		ERR_print_errors_fp(stderr);
+		SSL_CTX_free(_ctx[sock]);
+		_ctx[sock] = NULL;
 		return(-1);
 	}
 	
@@ -915,6 +949,9 @@ int SSLOpen(int sock, char *mycert, char *mykey, char *myproto, int certck, char
 	_check_fp = 1;
 	_check_digest = fingerprint;
 	_depth0ck = 0;
+	_firstrun = 1;
+	_verify_ok = 1;
+	_prev_err = -1;
 
 	if( mycert || mykey ) {
 
@@ -922,31 +959,51 @@ int SSLOpen(int sock, char *mycert, char *mykey, char *myproto, int certck, char
 	 * he does NOT have a separate certificate and private key file then
 	 * assume that it's a combined key and certificate file.
 	 */
+		char buffer[256];
+		
 		if( !mykey )
 			mykey = mycert;
 		if( !mycert )
 			mycert = mykey;
+
+		if ((!*remotename || !**remotename) && SSLCertGetCN(mycert, buffer, sizeof(buffer))) {
+			free(*remotename);
+			*remotename = xstrdup(buffer);
+		}
         	SSL_use_certificate_file(_ssl_context[sock], mycert, SSL_FILETYPE_PEM);
         	SSL_use_RSAPrivateKey_file(_ssl_context[sock], mykey, SSL_FILETYPE_PEM);
 	}
 
-	SSL_set_fd(_ssl_context[sock], sock);
-	
-	if(SSL_connect(_ssl_context[sock]) == -1) {
+	if (SSL_set_fd(_ssl_context[sock], sock) == 0 
+	    || SSL_connect(_ssl_context[sock]) < 1) {
 		ERR_print_errors_fp(stderr);
+		SSL_free( _ssl_context[sock] );
+		_ssl_context[sock] = NULL;
+		SSL_CTX_free(_ctx[sock]);
+		_ctx[sock] = NULL;
 		return(-1);
 	}
 
 	/* Paranoia: was the callback not called as we expected? */
-	if ((fingerprint != NULL || certck) && !_depth0ck) {
+	if (!_depth0ck) {
 		report(stderr, GT_("Certificate/fingerprint verification was somehow skipped!\n"));
-		
-		if( NULL != ( ssl = SSLGetContext( sock ) ) ) {
-			/* Clean up the SSL stack */
-			SSL_free( _ssl_context[sock] );
-			_ssl_context[sock] = NULL;
+
+		if (fingerprint != NULL || certck) {
+			if( NULL != SSLGetContext( sock ) ) {
+				/* Clean up the SSL stack */
+				SSL_shutdown( _ssl_context[sock] );
+				SSL_free( _ssl_context[sock] );
+				_ssl_context[sock] = NULL;
+				SSL_CTX_free(_ctx[sock]);
+				_ctx[sock] = NULL;
+			}
+			return(-1);
 		}
-		return(-1);
+	}
+
+	if (!certck && !fingerprint &&
+		(SSL_get_verify_result(_ssl_context[sock]) != X509_V_OK || !_verify_ok)) {
+		report(stderr, GT_("Warning: the connection is insecure, continuing anyways. (Better use --sslcertck!)\n"));
 	}
 
 	return(0);
@@ -957,39 +1014,16 @@ int SockClose(int sock)
 /* close a socket gracefully */
 {
 #ifdef	SSL_ENABLE
-    SSL *ssl;
-
-    if( NULL != ( ssl = SSLGetContext( sock ) ) ) {
+    if( NULL != SSLGetContext( sock ) ) {
         /* Clean up the SSL stack */
+        SSL_shutdown( _ssl_context[sock] );
         SSL_free( _ssl_context[sock] );
         _ssl_context[sock] = NULL;
+	SSL_CTX_free(_ctx[sock]);
+	_ctx[sock] = NULL;
     }
 #endif
 
-#ifdef __UNUSED__
-    /* 
-     * This hangs in RedHat 6.2 after fetchmail runs for a while a
-     * FIN_WAIT2 comes up in netstat and fetchmail never returns from
-     * the recv system call. (Reported from jtnews
-     * <jtnews@bellatlantic.net>, Wed, 24 May 2000 21:26:02.)
-     *
-     * Half-close the connection first so the other end gets notified.
-     *
-     * This stops sends but allows receives (effectively, it sends a
-     * TCP <FIN>).  */
-    if (shutdown(sock, 1) == 0) {
-	char ch;
-	/* If there is any data still waiting in the queue, discard it.
-	 * Call recv() until either it returns 0 (meaning we received a FIN)
-	 * or any error occurs.  This makes sure all data sent by the other
-	 * side is acknowledged at the TCP level.
-	 */
-	if (fm_peek(sock, &ch, 1) > 0)
-	    while (fm_read(sock, &ch, 1) > 0)
-		continue;
-    }
-#endif /* __UNUSED__ */
-
     /* if there's an error closing at this point, not much we can do */
     return(fm_close(sock));	/* this is guarded */
 }
@@ -1002,18 +1036,18 @@ int SockClose(int sock)
  */
 static ssize_t cygwin_read(int sock, void *buf, size_t count)
 {
-    char *bp = buf;
-    int n = 0;
+    char *bp = (char *)buf;
+    size_t n = 0;
 
-    if ((n = read(sock, bp, count)) == -1)
+    if ((n = read(sock, bp, count)) == (size_t)-1)
 	return(-1);
 
     if (n != count) {
-	int n2 = 0;
+	size_t n2 = 0;
 	if (outlevel >= O_VERBOSE)
 	    report(stdout, GT_("Cygwin socket read retry\n"));
 	n2 = read(sock, bp + n, count - n);
-	if (n2 == -1 || n + n2 != count) {
+	if (n2 == (size_t)-1 || n + n2 != count) {
 	    report(stderr, GT_("Cygwin socket read retry failed!\n"));
 	    return(-1);
 	}
@@ -1022,21 +1056,3 @@ static ssize_t cygwin_read(int sock, void *buf, size_t count)
     return count;
 }
 #endif /* __CYGWIN__ */
-
-#ifdef MAIN
-/*
- * Use the chargen service to test input buffering directly.
- * You may have to uncomment the `chargen' service description in your
- * inetd.conf (and then SIGHUP inetd) for this to work.  */
-main()
-{
-    int	 	sock = SockOpen("localhost", 19, NULL);
-    char	buf[80];
-
-    while (SockRead(sock, buf, sizeof(buf)-1))
-	SockWrite(1, buf, strlen(buf));
-    SockClose(sock);
-}
-#endif /* MAIN */
-
-/* socket.c ends here */
